import json

from tabulate import tabulate

from crawlab_ai import read_list, read_article
from crawlab_ai.utils.auth import get_token


def setup_crawl_parser(subparsers):
    crawl_parser = subparsers.add_parser("crawl", help="Crawl a website")
    crawl_parser.add_argument("url", help="URL to crawl")
    crawl_parser.add_argument(
        "-t",
        "--type",
        help="Type of the website to crawl",
        default="list",
        choices=["article", "list"],
    )
    crawl_parser.add_argument("-o", "--output", help="Output file path")
    crawl_parser.set_defaults(func=crawl)


def crawl(args):
    get_token()
    if args.type == "list":
        crawl_list(args)
    elif args.type == "article":
        crawl_article(args)


def crawl_list(args):
    df = read_list(args.url)
    if args.output:
        df.to_csv(args.output)
    else:
        print(tabulate(df, headers="keys", tablefmt="psql"))


def crawl_article(args):
    data = read_article(args.url)
    with open(args.output, "w") as f:
        f.write(json.dumps(data))



from crawlab_ai.utils.auth import get_token


def setup_gen_code_parser(subparsers):
    gen_code_parser = subparsers.add_parser(
        "gen_code", help="Generate crawler code for a webpage"
    )
    gen_code_parser.set_defaults(func=gen_code)


def gen_code(args):
    get_token()



from typing import Any, List, Iterable

import requests
from scrapy import Request
from scrapy.http import Response

from crawlab_ai.scrapy.base import BaseSpider
from crawlab_ai.utils.auth import get_auth_headers
from crawlab_ai.utils.logger import logger


class ScrapyListSpider(BaseSpider):
    """
    The ScrapyListSpider class is responsible for crawling a given URL and extracting data based on specified fields.
    It fetches rules from an API endpoint, which are used to identify list elements and fields within the webpage.
    The class also supports pagination by identifying the next page element and recursively crawling until no more pages
    are found.

    Attributes:
        fields (List[dict]): A list of fields to be extracted from each list element.

    Methods:
        parse(response: Response, **kwargs: Any) -> Iterable: Parses the response object and extracts data from the list
            elements.

    Usage:
        class MyListSpider(ScrapyListSpider):
            name = 'my_list_spider'
            start_urls = ['https://quotes.toscrape.com']
            fields = [
                {
                    'name': 'author',
                    'element_css_selector': '.author',
                    'is_text': True,
                },
                {
                    'name': 'quote',
                    'element_css_selector': '.text',
                    'is_text': True,
                },
                {
                    'name': 'tags',
                    'element_css_selector': '.tag',
                    'is_text': True,
                },
            ]

    """

    _fields: List[dict] = None
    _list_element_css_selector: str = None
    _next_page_element_css_selector: str = None

    fields: List[dict] = None

    def __init__(self):
        super(ScrapyListSpider, self).__init__()
        self._fetch_rules()

    def start_requests(self) -> Iterable[Request]:
        yield Request(self.start_urls[0], self.parse)

    def _fetch_rules(self):
        logger.info("Fetching rules for URL: " + self.start_urls[0])
        res = requests.post(
            url=self._api_endpoint + "/rules/list",
            headers=get_auth_headers(),
            json={
                "url": self.start_urls[0],
                "fields": self.fields,
            },
        )
        data = res.json()
        self._list_element_css_selector = data["model_list"][0]["list_model"][
            "list_element_css_selector"
        ]
        self._fields = data["model_list"][0]["list_model"]["fields"]
        self._next_page_element_css_selector = data["model_list"][0][
            "next_page_element_css_selector"
        ]
        logger.info("Rules fetched.")
        logger.info("List element CSS selector: " + self._list_element_css_selector)
        logger.info("Fields: " + str(self._fields))
        logger.info(
            "Next page element CSS selector: "
            + str(self._next_page_element_css_selector)
        )

    def parse(self, response: Response, **kwargs: Any) -> Any:
        list_items = response.css(self._list_element_css_selector)
        for item in list_items:
            data = {}
            for field in self._fields:
                name = field["name"]
                if field["is_text"]:
                    selector = field["element_css_selector"] + "::text"
                else:
                    selector = (
                        field["element_css_selector"]
                        + "::attr("
                        + field["attribute"]
                        + ")"
                    )
                value = item.css(selector).get()
                data[name] = value
            yield data

        if self._next_page_element_css_selector:
            next_page_element = response.css(self._next_page_element_css_selector)
            next_page_href = next_page_element.css("a::attr(href)").get()
            if next_page_href:
                yield response.follow(next_page_href, self.parse)


if __name__ == "__main__":
    # dry run the scrapy spider
    from scrapy.crawler import CrawlerProcess
    from scrapy.utils.project import get_project_settings

    class TestScrapyListSpider(ScrapyListSpider):
        name = "list_spider"
        start_urls = ["https://quotes.toscrape.com"]

    process = CrawlerProcess(get_project_settings())
    process.crawl(TestScrapyListSpider)
    process.start()



import requests

from crawlab_ai.spider.base import BaseSpider
from crawlab_ai.utils.auth import get_auth_headers
from crawlab_ai.utils.env import get_api_endpoint
from crawlab_ai.utils.logger import logger


class ArticleSpider(BaseSpider):
    def __init__(self, url: str, get_html=None):
        super().__init__(url, get_html)
        self.data: dict | None = None
        self.url = url
        self.get_html = get_html or self.crawl

    def fetch_rules(self):
        logger.info("Fetching rules for URL: " + self.url)
        res = requests.post(
            url=get_api_endpoint() + "/rules/article",
            headers=get_auth_headers(),
            json={
                "url": self.url,
            },
        )
        if res.status_code != 200:
            raise Exception("Failed to fetch rules for URL: " + self.url)

        self.data = res.json()
        logger.info("Rules fetched successfully for URL: " + self.url)
        logger.info("Title: %s", self.data.get("title"))
        logger.info("Author: %s", self.data.get("author"))
        logger.info("Publish date: %s", self.data.get("publish_date"))
        logger.info("Content: %s", self.data.get("content")[:200])
        logger.info("Title CSS selector: %s", self.data.get("title_css_selector"))
        logger.info("Author CSS selector: %s", self.data.get("author_css_selector"))
        logger.info(
            "Publish date CSS selector: %s", self.data.get("publish_date_css_selector")
        )
        logger.info("Content CSS selector: %s", self.data.get("content_css_selector"))

    def crawl(self):
        logger.info("Crawling URL: " + self.url)
        self.fetch_rules()
        logger.info("Crawling completed for URL: " + self.url)


def read_article(url: str, get_html=None):
    spider = ArticleSpider(url, get_html)
    spider.crawl()
    return spider.data


if __name__ == "__main__":
    data = read_article("https://www.36kr.com/p/2601845967059847")
    print(data)



import os
import sys


def get_token() -> str:
    # Get token from environment variable
    if "CRAWLAB_TOKEN" in os.environ:
        return os.environ["CRAWLAB_TOKEN"]

    # Get token from file
    token_file_path = os.path.join(os.path.expanduser("~"), ".crawlab", "token")
    if os.path.exists(token_file_path):
        with open(token_file_path, "r") as f:
            return f.read()

    # Token not found
    print("API token not found. Please set the CRAWLAB_TOKEN environment variable.")
    print(
        "If you don't have an API token, please sign up and apply at https://www.crawlab.cn/ai"
    )
    sys.exit(1)


def set_token(token: str):
    token_file_path = os.path.join(os.path.expanduser("~"), ".crawlab", "token")
    with open(token_file_path, "w") as f:
        f.write(token)


def get_auth_headers() -> dict:
    return {"Authorization": f"Bearer {get_token()}"}



from typing import List, Optional
from urllib.parse import urljoin

import requests
from bs4 import BeautifulSoup
from pandas import DataFrame
from concurrent.futures import ThreadPoolExecutor, as_completed

from crawlab_ai.spider.base import BaseSpider
from crawlab_ai.utils.auth import get_auth_headers
from crawlab_ai.utils.env import get_api_endpoint
from crawlab_ai.utils.logger import logger


class ListSpider(BaseSpider):
    """
    The ListSpider class is responsible for crawling a given URL and extracting data based on specified fields.
    It fetches rules from an API endpoint, which are used to identify list elements and fields within the webpage.
    The class also supports pagination by identifying the next page element and recursively crawling until no more pages
     are found.

    Attributes:
        url (str): The URL to be crawled.
        fields (List[dict]): A list of fields to be extracted from each list element.
        data (List[dict]): A list to store the extracted data from each list element.
        get_html (function): A function to fetch the HTML content of a webpage. Defaults to the _get_html method.

    Methods:
        crawl(url: str): Starts the crawling process for the given URL.
    """

    def __init__(self, url: str, fields: List[dict] = None, get_html=None):
        super().__init__(url, get_html)
        self.rules = None
        self.url = url
        self.fields = fields
        self.data = []
        self.fetch_rules()

    @property
    def list_element_css_selector(self):
        return self.rules["list_model"]["list_element_css_selector"]

    @property
    def item_fields(self):
        return self.rules["list_model"]["fields"]

    @property
    def next_page_element_css_selector(self):
        return self.rules["next_page_element_css_selector"]

    def fetch_rules(self):
        logger.info("Fetching rules for URL: " + self.url)
        res = requests.post(
            url=get_api_endpoint() + "/rules/list",
            headers=get_auth_headers(),
            json={
                "url": self.url,
                "fields": self.fields,
            },
        )
        data = res.json()
        if res.status_code != 200:
            raise Exception("Failed to fetch rules for URL: " + self.url)
        self.rules = data["model_list"][0]
        logger.info("Rules fetched successfully for URL: " + self.url)
        logger.info("List element CSS selector: " + self.list_element_css_selector)
        logger.info("Fields: " + str(self.item_fields))
        logger.info(
            "Next page element CSS selector: "
            + str(self.next_page_element_css_selector)
        )

    def crawl(self):
        futures = []
        with ThreadPoolExecutor(max_workers=5) as executor:
            futures.append(executor.submit(self._fetch_data, self.url))

            for future in as_completed(futures):
                next_page_url = future.result()
                while next_page_url:
                    future = executor.submit(self._fetch_data, next_page_url)
                    next_page_url = future.result()

    @staticmethod
    def _get_html(url):
        res = requests.get(url)
        return res.text

    def _fetch_data(self, url):
        logger.info("Crawling URL: " + url)
        html = self.get_html(url)
        soup = BeautifulSoup(html, "html.parser")
        list_items = soup.select(self.list_element_css_selector)
        for item in list_items:
            row = {}
            for field in self.item_fields:
                name = field["name"]
                field_element = item.select_one(field["element_css_selector"])
                if not field_element:
                    continue
                if field["is_text"]:
                    value = field_element.text.strip()
                else:
                    value = field_element.get(field["attribute"])
                row[name] = value
            self.data.append(row)
        logger.info("Fetched " + str(len(list_items)) + " items from URL: " + url)

        if self.next_page_element_css_selector:
            next_page_element = soup.select_one(self.next_page_element_css_selector)
            if next_page_element:
                logger.info("Next page found: " + next_page_element.get("href"))
                next_page_href = next_page_element.get("href")
                return urljoin(url, next_page_href)


def _get_fields(fields: List[str] | dict = None) -> Optional[List[str] | List[dict]]:
    if isinstance(fields, list):
        return fields
    elif isinstance(fields, dict):
        return [{"name": k, "description": v} for k, v in fields.items()]
    else:
        return None


def read_list(
    url: str, fields: List[str] | dict = None, get_html=None, as_dataframe=True
) -> DataFrame | List[dict]:
    """
    Reads a list of items from a webpage and returns a DataFrame.

    Args:
        url (str): The URL to be crawled.
        fields (List[str] | dict): A list of fields to be extracted from each list element.
        get_html (function): A function to fetch the HTML content of a webpage. Defaults to the requests library.
        as_dataframe (bool): Whether to return the extracted data as a DataFrame. Defaults to True.

    Returns:
        DataFrame | List[dict]: A DataFrame containing the extracted data if as_dataframe is True, otherwise a list of
        dictionaries.

    """
    spider = ListSpider(url=url, fields=_get_fields(fields), get_html=get_html)
    spider.crawl()
    if as_dataframe:
        return DataFrame(spider.data)
    else:
        return spider.data


if __name__ == "__main__":
    df = read_list("https://quotes.toscrape.com")
    print(df)
    # df = read_list('https://36kr.com/', [
    #     'title',
    #     'author',
    #     'url',
    # ])
    # print(df)



from crawlab_ai.utils.auth import set_token
from crawlab_ai.utils.env import set_api_endpoint


def setup_config_parser(subparsers):
    config_parser = subparsers.add_parser("config", help="Configure crawlab")
    config_parser.add_argument(
        "-u", "--url", help="Crawlab API URL", default="http://localhost:8000"
    )
    config_parser.add_argument("-t", "--token", help="Crawlab API token")
    config_parser.set_defaults(func=config)


def config(args):
    if args.token:
        set_token(args.token)

    if args.url:
        set_api_endpoint(args.url)



from setuptools import setup, find_packages

setup(
    name="crawlab-ai",
    version="0.0.4",
    packages=find_packages(),
    url="https://github.com/crawlab-team/crawlab-ai-sdk",
    license="MIT",
    author="Marvin Zhang",
    author_email="tikazyq@163.com",
    description="SDK for Crawlab AI",
    package_dir={"crawlab_ai": "crawlab_ai"},
    keywords=["crawlab", "ai"],
    entry_points={
        "console_scripts": [
            "crawlab-ai=crawlab_ai.cli:main",
        ]
    },
)
